#ifndef MATRIX3X3_H
#define MATRIX3X3_H

#include <stdio.h>
#include "point3d.h"

typedef struct {
	/* Indexes representing the matrix like this: 
	 *
	 * [0 1 2]
	 * [3 4 5]
	 * [6 7 8] 
	 *
	 * */
	float data[9];
} Matrix3x3;

typedef struct {
	/* Indexes representing the matrix like this: 
	 *
	 * [0  1  2  3 ]
	 * [4  5  6  7 ]
	 * [8  9  10 11]
	 * [12 13 14 15]
	 *
	 * */
	float data[16];
} Matrix4x4;

inline Matrix3x3 matrix3x3_create() {
	Matrix3x3 result;

	/* Compiler would probably optimalize this on its own */
	result.data[0] = 0;
	result.data[1] = 0;
	result.data[2] = 0;
	result.data[3] = 0;
	result.data[4] = 0;
	result.data[5] = 0;
	result.data[7] = 0;
	result.data[8] = 0;

	return result;
}

inline Matrix3x3 matrix3x3_create_from_array(float* arr) {
	Matrix3x3 result;

	for(int i = 0; i < 9; ++i) {
		result.data[i] = arr[i];
	}

	return result;
}

inline void print_matrix3x3(Matrix3x3 m) {
	printf("| %.12f %.12f %.12f |\n", m.data[0], m.data[1], m.data[2]);
	printf("| %.12f %.12f %.12f |\n", m.data[3], m.data[4], m.data[5]);
	printf("| %.12f %.12f %.12f |\n", m.data[6], m.data[7], m.data[8]);
}

inline Matrix3x3 matrix3x3_matrices_multiplication(Matrix3x3 m1, Matrix3x3 m2) {
	Matrix3x3 result = matrix3x3_create();

	for(int col = 0; col < 3; ++col) {
		for(int row = 0; row < 3; ++row) {
			for(int i = 0; i < 3; ++i) {
				result.data[col * 3 + row] += m1.data[col * 3 + i] * m2.data[row + 3 * i];
			}
		}
	}

	return result;
}

inline Point3D matrix3x3_vector_multiply(Matrix3x3 m, Point3D vector) {
	float res[3];
	for(int i = 0; i < 3; ++i) {
		res[i] = m.data[3 * i] * vector.x + m.data[3 * i + 1] * vector.y + m.data[3 * i + 2] * vector.z;
	}

	return create_point3d(res[0], res[1], res[2]);
}

inline Point3D matrix3x3_vector_multiply_data(float* data, Point3D vector) {
	float res[3];
	for(int i = 0; i < 3; ++i) {
		res[i] = data[3 * i] * vector.x + data[3 * i + 1] * vector.y + data[3 * i + 2] * vector.z;
	}

	return create_point3d(res[0], res[1], res[2]);
}

inline void matrix3x3_put_to_column(Matrix3x3* m, Point3D vector, int i) {
	m->data[i] = vector.x;
	m->data[i + 3] = vector.y;
	m->data[i + 6] = vector.z;
}

inline void matrix3x3_put_to_row(Matrix3x3* m, Point3D vector, int i) {
	m->data[i * 3] = vector.x;
	m->data[i * 3 + 1] = vector.y;
	m->data[i * 3 + 2] = vector.z;
}

inline Matrix4x4 matrix4x4_create() {
	Matrix4x4 result;

	for(int i = 0; i < 16; ++i) {
		result.data[i] = 0;
	}

	return result;
}

inline Matrix4x4 matrix4x4_transpose(Matrix4x4 m) {
	Matrix4x4 result;

	result.data[0] = m.data[0];
	result.data[1] = m.data[4];
	result.data[2] = m.data[8];
	result.data[3] = m.data[12];

	result.data[4] = m.data[1];
	result.data[5] = m.data[5];
	result.data[6] = m.data[9];
	result.data[7] = m.data[13];

	result.data[8] = m.data[2];
	result.data[9] = m.data[6];
	result.data[10] = m.data[10];
	result.data[11] = m.data[14];

	result.data[12] = m.data[3];
	result.data[13] = m.data[7];
	result.data[14] = m.data[11];
	result.data[15] = m.data[15];

	return result;
}

inline Matrix4x4 matrix4x4_matrices_multiplication(Matrix4x4 m1, Matrix4x4 m2) {
	Matrix4x4 result = matrix4x4_create();

	for(int col = 0; col < 4; ++col) {
		for(int row = 0; row < 4; ++row) {
			for(int i = 0; i < 4; ++i) {
				result.data[col * 4 + row] += m1.data[col * 4 + i] * m2.data[row + 4 * i];
			}
		}
	}

	return result;
}

inline void matrix4x4_put_to_row(Matrix4x4* m, Point3D vector, int i) {
	m->data[i * 4] = vector.x;
	m->data[i * 4 + 1] = vector.y;
	m->data[i * 4 + 2] = vector.z;
}

//Matrix3x3 matrix3x3_create();
//Matrix3x3 matrix3x3_create_from_array(float* arr);
//Matrix3x3 matrix3x3_matrices_multiplication(Matrix3x3 m1, Matrix3x3 m2);
//Point3D matrix3x3_vector_multiply(Matrix3x3 m, Point3D vector);
//Point3D matrix3x3_vector_multiply_data(float* m, Point3D vector);
//
//void matrix3x3_put_to_column(Matrix3x3* m, Point3D vector, int i);
//void matrix3x3_put_to_row(Matrix3x3* m, Point3D vector, int i);
//
//void print_matrix3x3(Matrix3x3 m);
//
//Matrix4x4 matrix4x4_create();
//Matrix4x4 matrix4x4_transpose(Matrix4x4 m);
//Matrix4x4 matrix4x4_matrices_multiplication(Matrix4x4 m1, Matrix4x4 m2);
//void matrix4x4_put_to_row(Matrix4x4* m, Point3D vector, int i);

#endif
