#include <stdio.h>
#include "transform3d.h"
#include <stdlib.h>
#include "time_measurement.h"
#include "math_utils.h"
#include "cluster_registration.h"
#include <time.h>
#include "hashtable.h"
#include "point3d.h"
#include "proximity_verifier.h"
#include "config.h"
#include "density_clusterer.h"

#include "sampler.h"

Matrix3x3 find_rotation(Point3D* set, int size) {
	Matrix3x3 rot = matrix3x3_create();
	Matrix3x3 m = matrix3x3_create();

	for(int i = 0; i < size; ++i) {
		m.data[0] += set[i].x * set[i].x;
		m.data[1] += set[i].x * set[i].y;
		m.data[2] += set[i].x * set[i].z;
		m.data[4] += set[i].y * set[i].y;
		m.data[5] += set[i].y * set[i].z;
		m.data[8] += set[i].z * set[i].z;
	}

	m.data[3] = m.data[1];
	m.data[6] = m.data[2];
	m.data[7] = m.data[5];


	Matrix3x3 eigens;

	/* Beware! Might return different order of eigenvectors than reference registrator!
	 * (However when tested on same data it works). */
	evd(&m, &eigens);

	/*  Reference data: 
	 	m.data[1] = 0.457117;
		m.data[2] = -0.0475125;
		m.data[3] = -0.00641582;
		m.data[3] = -0.0475125;
		m.data[4] = 0.95353;
		m.data[5] = 0.262307;
		m.data[6] = -0.00641582;
		m.data[7] = 0.262307;
		m.data[8] = 0.122265;

		Reference result:
		DenseMatrix 3x3-Double
			-0.0172174    0.99648  -0.0820406
			-0.278542  0.0740241    0.957567
			0.96027  0.0393385    0.276287
		*/

	//eigens.data[0] = -0.00670258;
	//eigens.data[1] = 0.994209;
	//eigens.data[2] = -0.1072;
	//eigens.data[3] = -0.277431;
	//eigens.data[4] = 0.101197;
	//eigens.data[5] = 0.955401;
	//eigens.data[6] =  0.960722;
	//eigens.data[7] =  0.0361592;
	//eigens.data[8] =  0.275146;

	/* rx is 1st row of eigenvalue matrix */
	Point3D rx = create_point3d(eigens.data[0], eigens.data[1], eigens.data[2]);
	/* ry is 2nd row */
	Point3D ry = create_point3d(eigens.data[3], eigens.data[4], eigens.data[5]);
	Point3D rz = POINTS_CROSSPRODUCT(rx, ry);

	matrix3x3_put_to_column(&rot, rx, 0);
	matrix3x3_put_to_column(&rot, ry, 1);
	matrix3x3_put_to_column(&rot, rz, 2);

	return rot;
}

Transform3D build_best_candidate(Transform3D decorrelation, Transform3D best_match) {
	Point3D rotated_translation = create_point3d(decorrelation.translation[0], decorrelation.translation[1], decorrelation.translation[2]);
	transform3d_rotate_point_ref(decorrelation, &rotated_translation);
	float data[] = {rotated_translation.x, rotated_translation.y, rotated_translation.z};

	Transform3D rott = transform3d_create(decorrelation.rotation, data);
	Transform3D result;
	memset(result.rotation, '\0', sizeof(float) * 9);

	for(int col = 0; col < 3; ++col) {
		for(int row = 0; row < 3; ++row) {
			for(int i = 0; i < 3; ++i) {
				result.rotation[col * 3 + row] += best_match.rotation[col * 3 + i] * rott.rotation[row + 3 * i];
			}
		}
	}

	for(int i = 0; i < 3; ++i) {
		result.translation[i] = best_match.rotation[i * 3] * rott.translation[0];
		result.translation[i] += best_match.rotation[i * 3 + 1] * rott.translation[1];
		result.translation[i] += best_match.rotation[i * 3 + 2] * rott.translation[2];
		result.translation[i] += best_match.translation[i];
	}

	return result;
}

Transform3D cluster_registration_execute(TriangleMesh* restrict mesh1, TriangleMesh* restrict mesh2) {

	trianglemesh_init_radiuses(mesh2);

	Point3D to_center = mesh2->centroid;
	point_negative_ref(&to_center);
	/* Center mesh so its pointsum is zero */
	for(int i = 0; i < mesh2->points_size; ++i) {
		POINTS_ADD_REF(&(mesh2->points[i]), to_center);
	}

	Matrix3x3 rotation = find_rotation(mesh2->points, mesh2->points_size);

	float to_center_arr[] = { to_center.x, to_center.y, to_center.z };
	Transform3D decorrelating_transformation = transform3d_create(rotation.data, to_center_arr);

	/* Transform mesh2 so its points are uncorrelated */
	for(int i = 0; i < mesh2->points_size; ++i) {
		transform3d_rotate_point_ref(decorrelating_transformation, &(mesh2->points[i]));
		transform3d_rotate_point_ref(decorrelating_transformation, &(mesh2->normals[i]));
	}

	metric_data_init(mesh2, mesh2->area_sum, mesh2->weighted_radius, mesh2->radius, config_rotation_coeff);

	/* Initialization of needed datastructures */
	ProximityVerifier* verifier = create_proximity_verifier(mesh1, mesh2, DELTA);

	int candidates_size;
	Transform3D* candidates = curvature_oracle_find_candidates(mesh1, mesh2, verifier, &candidates_size);

	density_clusterer_init(1, mesh2->radius);

	Transform3D best_candidate = density_clusterer_get_best_candidate(candidates, candidates_size);

	proximity_verifier_free(verifier);
	free(candidates);

	/* The final transformation consist of applying decorrelation transformation
	 * and the best found transformation. */
	return build_best_candidate(decorrelating_transformation, best_candidate);
}
