#include <stdio.h>
#include "corner_table.h"

int edge_hash(Edge e) {
	return (e.v1 + e.v2) % DICTIONARY_SIZE;
}

DictionaryCTNode* dictionary_ct_node_create(int index, Edge edge) {
	DictionaryCTNode* node = malloc(sizeof(DictionaryCTNode));

	node->index = index;
	node->edge = edge;
	node->next = NULL;

	return node;
}

void dictionary_ct_free(DictionaryCT* dic) {
	if(dic == NULL) {
		return;
	}

	for(int i = 0; i < DICTIONARY_SIZE; ++i) {
		DictionaryCTNode* node = dic->nodes[i];
		DictionaryCTNode* temp;
		while(node != NULL) {
			temp = node->next;
			free(node);
			node = temp;
		}
	}

	free(dic);
}

int dictionary_ct_get(DictionaryCT* dic, Edge e) {
	int hash = edge_hash(e);

	DictionaryCTNode* node = dic->nodes[hash];
	while(node != NULL) {
		Edge node_edge = node->edge;
		/* If edges have same values for both vertices, they're equal,
		 * therefore dictionary contains edge 'e' */
		if(node_edge.v1 == e.v1 && node_edge.v2 == e.v2) {
			return node->index;
		}
		node = node->next;
	}

	return CT_DIC_NODATA;
}

//int dictionary_ct_contains(dictionary_ct* dic, edge e) {
//	int hash = edge_hash(e);
//
//	dictionary_ct_node* node = dic->nodes[hash];
//
//	while(node != NULL) {
//		edge node_edge = node->edge;
//		/* If edges have same values for both vertices, they're equal,
//		 * therefore dictionary contains edge 'e' */
//		if(node_edge.v1 == e.v1 && node_edge.v2 == e.v2) {
//			return 1;
//		}
//		node = node->next;
//	}
//
//	return 0;
//}

void dictionary_ct_add(DictionaryCT* dic, int index, Edge edge) {
	int hash = edge_hash(edge);

	DictionaryCTNode* node = dic->nodes[hash];
	DictionaryCTNode* new_node = dictionary_ct_node_create(index, edge);

	/* There already are some data for 'hash' */
	if(node != NULL) {
		new_node->next = node;
	} 

	dic->nodes[hash] = new_node;
}

DictionaryCT* dictionary_ct_create() {
	DictionaryCT* dic = malloc(sizeof(DictionaryCT));
	memset(dic->nodes, '\0', sizeof(DictionaryCTNode*) * DICTIONARY_SIZE);

	return dic;
}

void corner_table_process_edge(int* opposite, DictionaryCT* dic, Edge e, int index) {
	int opposite_corner_index = dictionary_ct_get(dic, e);

	/* Edge was already seen once */
	if(opposite_corner_index != CT_DIC_NODATA) {
		opposite[index] = opposite_corner_index;
		opposite[opposite_corner_index] = index;
	} else {
		/* Dictionary doesn't contain edge 'e' yet */
		Edge reversed_edge = {e.v2, e.v1};
		/* TODO: Non-Manifold */
		dictionary_ct_add(dic, index, reversed_edge);
	}
}

int* corner_table_neighbours(CornerTable* ct, int vertex, int* result_length) {

	VECTOR_CREATE(int) vector;
	VECTOR_INIT(vector, int);
	//VectorInt* vector = vector_int_create();

	int initial_corner = ct->incident_corner[vertex];
	VECTOR_ADD(vector, corner_table_get_vertex(ct, corner_table_prev_corner(ct, initial_corner)));
	//vector_int_add(vector, corner_table_get_vertex(ct, corner_table_prev_corner(ct, initial_corner)));

	int o = ct->opposite[corner_table_next_corner(ct, initial_corner)];

	while(o != CT_NODATA) {
		/* Loop ends the moment it gets back to the initial corner, that
		 * means when it will make whole loop around the vertex */
		if(corner_table_next_corner(ct, o) == initial_corner) {
			break;
		}

		VECTOR_ADD(vector, corner_table_get_vertex(ct, o));
		//vector_int_add(vector, corner_table_get_vertex(ct, o));
		o = ct->opposite[corner_table_prev_corner(ct, o)];
	}

	/* In case this vertex is at border */
	if(o == CT_NODATA) {

		/* Loop around the vertex in opposite direction.
		 * That's instead of looping through 'prev' corners loop
		 * via 'next' corners instead. */
		VECTOR_ADD(vector, corner_table_get_vertex(ct, corner_table_next_corner(ct, initial_corner)));
		//vector_int_add(vector, corner_table_get_vertex(ct, corner_table_next_corner(ct, initial_corner)));
		o = ct->opposite[corner_table_prev_corner(ct, initial_corner)];

		while(o != CT_NODATA) {
			if(corner_table_prev_corner(ct, o) == initial_corner) {
				break;
			}

			VECTOR_ADD(vector, corner_table_get_vertex(ct, o));
			//vector_int_add(vector, corner_table_get_vertex(ct, o));
			o = ct->opposite[corner_table_next_corner(ct, o)];
		}
	}

	*result_length = vector.size;
	return vector.data;
}

CornerTable* corner_table_create(const TriangleMesh* mesh) {
	CornerTable* ct = calloc(1, sizeof(CornerTable));

	ct->opposite = malloc(sizeof(int) * mesh->triangles_size * 3);
	ct->incident_corner = malloc(sizeof(int) * mesh->points_size);
	ct->mesh = mesh;

	for(int i = 0; i < mesh->triangles_size * 3; ++i) {
		ct->opposite[i] = CT_NODATA;
	}

	DictionaryCT* dictionary = dictionary_ct_create();

	Edge e;
	int corner_index;
	for(int i = 0; i < mesh->triangles_size; ++i) {
		Triangle t = mesh->triangles[i];

		e.v1 = t.v1;
		e.v2 = t.v2;
		corner_index = 3 * i;
		corner_table_process_edge(ct->opposite, dictionary, e, corner_index);
		ct->incident_corner[t.v3] = corner_index;

		e.v1 = t.v2;
		e.v2 = t.v3;
		corner_index = 3 * i + 1;
		corner_table_process_edge(ct->opposite, dictionary, e, corner_index);
		ct->incident_corner[t.v1] = corner_index;

		e.v1 = t.v3;
		e.v2 = t.v1;
		corner_index = 3 * i + 2;
		corner_table_process_edge(ct->opposite, dictionary, e, corner_index);
		ct->incident_corner[t.v2] = corner_index;

	}

	dictionary_ct_free(dictionary);

	return ct;
}

int corner_table_next_corner(CornerTable* ct, int corner_index) {
	if(corner_index % 3 == 2) {
		return corner_index - 2;
	} else {
		return corner_index + 1;
	}
}

int corner_table_prev_corner(CornerTable* ct, int corner_index) {
	if(corner_index % 3 == 0) {
		return corner_index + 2;
	} else {
		return corner_index - 1;
	}
}

int corner_table_get_vertex(CornerTable* ct, int corner_index) {
	int triangle_index = corner_index / 3;

	if(corner_index % 3 == 0) {
		return ct->mesh->triangles[triangle_index].v3;
	}
	if(corner_index % 3 == 1) {
		return ct->mesh->triangles[triangle_index].v1;
	}
	return ct->mesh->triangles[triangle_index].v2;
}

void corner_table_free(CornerTable* ct) {
	free(ct->opposite);
	free(ct->incident_corner);
	free(ct);
}
