
#include <time.h>
#include <stdio.h>

#include "curvature_oracle.h"

#include "time_measurement.h"

int* curvature_oracle_get_neighbours_euclid(int vertex_index, const NeighboursCache* neighbours_cache, float dist, const TriangleMesh* mesh, int* result_length) {

	VECTOR_CREATE(int) vector;
	VECTOR_INIT(vector, int);

	VECTOR_CREATE(int) stack;
	VECTOR_INIT(stack, int);

	//StackInt* stack = stack_int_create();
	char* visited = calloc(mesh->points_size, sizeof(char));

	float dist_squared = dist * dist;

	visited[vertex_index] = 1;
	VECTOR_ADD(stack, vertex_index);
	VECTOR_ADD(vector, vertex_index);

	Point3D center = mesh->points[vertex_index];

	/* Iterate while there still are items in the stack */
	while(stack.size) {
		int curr_vertex = stack.data[stack.size - 1];
		stack.size -= 1;
		
		int neighbours_size = 0;
		int* neighbours = neighbours_cache_get(neighbours_cache, curr_vertex, &neighbours_size);

		/* Iterate over neighbours for curr_vertex */
		for(int i = 0; i < neighbours_size; ++i) {
			if(visited[neighbours[i]]) {
				continue;
			}
			/* If neighbour satisfy the condition of being 'dist' units away. */
			if(points_distance_squared(mesh->points[neighbours[i]], center) <= dist_squared) {
				/* Add it to the result */
				VECTOR_ADD(vector, neighbours[i]);

				/* It might have other points 'dist' units away in its neighbourhood. 
				 * Therefore add them to stack to be processed */
				VECTOR_ADD(stack, neighbours[i]);

				visited[neighbours[i]] = 1;
			}
		}
	}

	free(visited);
	VECTOR_FREE(stack);

	*result_length = vector.size;
	return vector.data;
}

int is_singular3x3(float* m) {

	float D = m[0] * m[2] - m[1] * m[1];

	if (fabs(D) < 0.0000000001) {
		return 1;
	}

	float D2 = m[0] * m[3];
	float y = D2 / D;

	if (fabs(m[3] * y - m[4]) < 0.000001) {
		return 1;
	}

	return 0;
}

void evd2x2(float* e1, float* e2, float* v1, float* v2, float a, float b, float c) {
	float d = sqrt((a - c) * (a - c) + 4 * b * b);
	*e1 = (a + c + d) / 2;
	*e2 = (a + c - d) / 2;

	if (fabs(*e2) < fabs(*e1)) {
		float tmp = *e2;
		*e2 = *e1;
		*e1 = tmp;
	}

	*v1 = 1;
	*v2 = 1;
	if (fabs(b) < fabs(a - *e1)) {
		*v1 = -b / (a - *e1);
	}
	else {
		*v2 = (*e1 - a) / b;
	}
}

void solve_symetric_stripe3x3(float a, float b, float c, float d, float e, float r[], float result[]) {
	float det = a * c * e - d * d * a - e * b * b;
	float det1 = r[0] * c * e + b * d * r[2] - d * d * r[0] - e * r[1] * b;
	float det2 = a * r[1] * e - r[2] * d * a - e * b * r[0];
	float det3 = a * c * r[2] + r[0] * b * d - d * r[1] * a - r[2] * b * b;

	result[0] = det1 / det;
	result[1] = det2 / det;
	result[2] = det3 / det;
}

int curvature_oracle_curvature_equality_condition(PointRecord p) {
	float ratio = p.k1 / p.k2;
	if(ratio < config_curvature_ratio) {
		return 1;
	}

	if(ratio > (1 / config_curvature_ratio)) {
		return 1;
	}

	return 0;
}

/* Returns:
 * 		1 - If it's possible to calculate curvatures
 * 		0 - If something fails.
 */
int curvature_oracle_compute_curvature(int vertex_index, const NeighboursCache* neighbours_cache, const TriangleMesh* mesh, PointRecord* result) {
	int neighbours_size = 0;
	int* neighbours = curvature_oracle_get_neighbours_euclid(vertex_index, neighbours_cache, mesh->avgEL * config_neighbourhood_width, mesh, &neighbours_size);
	
	if(neighbours == NULL || neighbours_size == 0) {
		return 0;
	}

	Point3D point = mesh->points[vertex_index];
	Point3D v3 = mesh->normals[vertex_index];
	Point3D v1 = POINTS_SUB(mesh->points[neighbours[1]], point);

	/* v1 -= v3 * (v1 . v3) */
	POINTS_SUB_REF(&v1, POINT_SCALE(v3, POINTS_DOTPRODUCT(v1, v3)));

	if(POINT_ABS(v1) < 0.000001) {
		free(neighbours);
		return 0;
	}

	POINT_SCALE_REF(&v1, 1.0 / POINT_ABS(v1));

	Point3D v2 = POINTS_CROSSPRODUCT(v1, v3);

	/* Matrix of rotation, we only need forst two rows */
	float rot[6];
	rot[0] = v1.x;
	rot[3] = v2.x;

	rot[1] = v1.y;
	rot[4] = v2.y;

	rot[2] = v1.z;
	rot[5] = v2.z;

	/* Rotated translation */
	float rtr[] = {
		-(rot[0] * point.x) - (rot[1] * point.y) - (rot[2] * point.z),
		-(rot[3] * point.x) - (rot[4] * point.y) - (rot[5] * point.z)
	};

	float m[5] = {0};
	float r[3] = {0};

	for(int i = 0; i < neighbours_size; ++i) {
		int point_index = neighbours[i];

		float w = mesh->weights[point_index];
		Point3D p = mesh->points[point_index];
		Point3D normal = mesh->normals[point_index];

		float lpx = rot[0] * p.x + rot[1] * p.y + rot[2] * p.z + rtr[0];
		float lpy = rot[3] * p.x + rot[4] * p.y + rot[5] * p.z + rtr[1];
		float lnx = rot[0] * normal.x + rot[1] * normal.y + rot[2] * normal.z;
		float lny = rot[3] * normal.x + rot[4] * normal.y + rot[5] * normal.z;

		float wlpx = w * lpx;
		float wlpy = w * lpy;

		float wlpxlpx = wlpx * lpx;
		float wlpxlpy = wlpx * lpy;
		float wlpylpy = wlpy * lpy;


		m[0] += wlpxlpx;
		m[1] += wlpxlpy;
		m[2] += wlpylpy;
		r[0] += wlpx * lnx;
		r[1] += wlpy * lnx;

		m[2] += wlpxlpx;
		m[3] += wlpxlpy;
		m[4] += wlpylpy;
		r[1] += wlpx * lny;
		r[2] += wlpy * lny;
	}

	free(neighbours);

	if(is_singular3x3(m)) {
		return 0;
	}

	float s[3];
	solve_symetric_stripe3x3(m[0], m[1], m[2], m[3], m[4], r, s);

	float e1;
	float e2;
	float egv1, egv2;

	evd2x2(&e1, &e2, &egv1, &egv2, s[0], s[1], s[2]);

	POINT_SCALE_REF(&v1, egv1);
	POINT_SCALE_REF(&v2, egv2);
	POINTS_ADD_REF(&v1, v2);
	POINT_SCALE_REF(&v1, 1 / POINT_ABS(v1));

	PointRecord point_record = point_record_create(e1, e2, v1, v3, point);
	*result = point_record;

	return 1;
}

/* Worker function - computes curvatures in specified range */
void* curvature_oracle_calculate_curvature_worker(void* arguments) {
	CurvatureOraclePRArgs args = *((CurvatureOraclePRArgs*) arguments);

	int start = args.start;
	int n = args.n;
	PointRecord* points = args.points;
	int points_size = args.points_size;
	const NeighboursCache* cache = args.cache;
	const TriangleMesh* mesh = args.mesh;
	int* samples = args.samples;

	if(start + n >= args.points_size) {
		n -= (start + n) - args.points_size;
	}

	for(int i = start; i < start + n; ++i) {
		PointRecord point_record;
		int ret = curvature_oracle_compute_curvature(samples[i], cache, mesh, &point_record);
		points[i] = point_record;
		/* Computation of curvature failed */
		if(ret == 0) {
			points[i].k1 = NAN;
		}
	}
	return NULL;
}

/* Prepares data for workers and starts them */
PointRecord* curvature_oracle_start_curvature_workers(const TriangleMesh* mesh, int* sampled_mesh, int sampled_mesh_size) {

	PointRecord* point_records1 = malloc(sizeof(PointRecord) * sampled_mesh_size);
	CornerTable* ctable1 = corner_table_create(mesh);
	NeighboursCache* cache1 = neighbours_cache_create(mesh->points_size, ctable1);

	int number_of_processors = config_number_of_processors;
	int to_process = ceil((float) sampled_mesh_size / number_of_processors);

	pthread_t tid;
	pthread_t* tids = malloc(sizeof(pthread_t) * number_of_processors);
	CurvatureOraclePRArgs** t_arg_pointers = malloc(sizeof(CurvatureOraclePRArgs*) * number_of_processors);

	for(int i = 0; i < number_of_processors; ++i) {
		CurvatureOraclePRArgs* ta = malloc(sizeof(CurvatureOraclePRArgs));
		ta->start = i * to_process;
		ta->points = point_records1;
		ta->points_size = sampled_mesh_size;
		ta->n = to_process;
		ta->samples = sampled_mesh;
		ta->cache = cache1;
		ta->mesh = mesh;

		pthread_create(&tid, NULL, curvature_oracle_calculate_curvature_worker, (void*) ta);

		tids[i] = tid;
		t_arg_pointers[i] = ta;
	}

	for(int i = 0; i < number_of_processors; ++i) {
		pthread_join(tids[i], NULL);
	}

	for(int i = 0; i < number_of_processors; ++i) {
		free(t_arg_pointers[i]);
	}

	free(tids);
	free(t_arg_pointers);
	corner_table_free(ctable1);
	neighbours_cache_free(&cache1);

	return point_records1;
}


static pthread_mutex_t candidates_mutex;

void* curvature_oracle_find_matches_worker(void* arguments) {
	CurvatureOracleFMArgs args = *((CurvatureOracleFMArgs*) arguments);

	int start = args.start;
	int n = args.n;
	K2Tree* tree = args.tree;
	const NeighboursCache* cache = args.cache;
	const TriangleMesh* mesh = args.mesh;
	const ProximityVerifier* verifier = args.verifier;
	CandidatesTest* ct = args.ct;

	if(start + n >= args.candidates_size) {
		n -= (start + n) - args.candidates_size;
	}

	int i = start;
	for(i = start; i < start + n; ++i) {

		PointRecord point_record;

		int curvature_success = curvature_oracle_compute_curvature(i, cache, mesh, &point_record);
		/* Computation of curvature failed */
		if(!curvature_success || !curvature_oracle_curvature_equality_condition(point_record)) {
			continue;
		}

		/* Find point from the other mesh with nearest curvatures to this one */
		PointRecord nearest = k2tree_find_nearest(tree, point_record);

		/* Construct transformation mapping points to each other */
		Transform3D transformation = transform3d_create_map(point_record, nearest, 0);
		Transform3D transformation_swapped = transform3d_create_map(point_record, nearest, 1);

		Transform3D transforms_to_test[] = {transformation, transformation_swapped};

		for(int j = 0; j < 2; ++j) {

			int verified = proximity_verifier_verify(verifier, transforms_to_test[j]);

			/* If transformation is verified as valid, add it to valid candidates */
			if(verified) {
				pthread_mutex_lock(&candidates_mutex);
				
				ct->candidates[ct->size] = transforms_to_test[j];
				ct->size += 1;

				pthread_mutex_unlock(&candidates_mutex);
			}
		}

		/* If enough candidates were generated - end the thread
		 * Caution, this might result in more candidates being return than specified
		 * in config. */
		if(ct->size >= config_candidates_length) {
			return NULL;
		}
	}

	return NULL;
}

CandidatesTest* curvature_oracle_start_find_matches_workers(const TriangleMesh* mesh, const ProximityVerifier* verifier, int* sampled_mesh, int sampled_mesh_size, PointRecord* curvatures, int curvatures_size) {
	K2Tree* tree = k2tree_create(curvatures, curvatures_size);

	CornerTable* ctable = corner_table_create(mesh);
	NeighboursCache* cache = neighbours_cache_create(mesh->points_size, ctable);

	CandidatesTest* ct = malloc(sizeof(CandidatesTest));
	ct->size = 0;
	ct->capacity = 2 * sampled_mesh_size;
	ct->candidates = malloc(sizeof(Transform3D) * ct->capacity);

	int number_of_processors = config_number_of_processors;
	int to_process = ceilf((float) sampled_mesh_size / number_of_processors);

	pthread_t tid;
	pthread_t* tids = malloc(sizeof(pthread_t) * number_of_processors);
	CurvatureOracleFMArgs** t_arg_pointers = malloc(sizeof(CurvatureOracleFMArgs*) * number_of_processors);

	for(int i = 0; i < number_of_processors; ++i) {
		CurvatureOracleFMArgs* ta = malloc(sizeof(CurvatureOracleFMArgs));
		ta->start = i * to_process;
		ta->tree = tree;
		ta->n = to_process;
		ta->cache = cache;
		ta->mesh = mesh;
		ta->verifier = verifier;
		ta->candidates_size = sampled_mesh_size;
		ta->ct = ct;

		pthread_create(&tid, NULL, curvature_oracle_find_matches_worker, (void*) ta);

		tids[i] = tid;
		t_arg_pointers[i] = ta;
	}

	for(int i = 0; i < number_of_processors; ++i) {
		pthread_join(tids[i], NULL);
	}

	for(int i = 0; i < number_of_processors; ++i) {
		free(t_arg_pointers[i]);
	}

	free(tids);
	free(t_arg_pointers);
	corner_table_free(ctable);
	neighbours_cache_free(&cache);
	k2tree_free(tree);

	return ct;

}

void curvature_oracle_sample_meshes(TriangleMesh* restrict mesh1, TriangleMesh* restrict mesh2, int** sampled_mesh1, int* sampled_mesh1_size, int** sampled_mesh2, int* sampled_mesh2_size) {
	sampler_uniform_sampling_grid(mesh1, config_candidates_length, (mesh2->radius * config_delta_ratio) / 3.0, sampled_mesh1, sampled_mesh1_size);

	*sampled_mesh2 = malloc(sizeof(int) * mesh2->points_size);
	for(int i = 0; i < mesh2->points_size; ++i) {
		(*sampled_mesh2)[i] = i;
	}
	*sampled_mesh2_size = mesh2->points_size;
}

Transform3D* curvature_oracle_find_candidates(TriangleMesh* restrict mesh1, TriangleMesh* restrict mesh2, ProximityVerifier* verifier, int* candidates_size) {
	int* sampled_mesh1;
	int sampled_mesh1_size;

	int* sampled_mesh2;
	int sampled_mesh2_size;

	curvature_oracle_sample_meshes(mesh1, mesh2, &sampled_mesh1, &sampled_mesh1_size, &sampled_mesh2, &sampled_mesh2_size);
	PointRecord* r = curvature_oracle_start_curvature_workers(mesh1, sampled_mesh1, sampled_mesh1_size);

	PointRecord* curvatures = malloc(sizeof(PointRecord) * sampled_mesh1_size);
	int curvatures_size = 0;

	/* TODO: This is kinda awkward - make this filtering process
	 * part of calculating the curvature analogously to how candidates 
	 * are being find */
	/* Removing invalid curvatures */
	for(int i = 0; i < sampled_mesh1_size; ++i) {
		if(isnan(r[i].k1) || !curvature_oracle_curvature_equality_condition(r[i])) {
			continue;
		}

		curvatures[curvatures_size] = r[i];
		curvatures_size += 1;
	}

	free(r);
	free(sampled_mesh1);

	CandidatesTest* ct = curvature_oracle_start_find_matches_workers(mesh2, verifier, sampled_mesh2, sampled_mesh2_size, curvatures, curvatures_size);

	free(curvatures);
	free(sampled_mesh2);

	*candidates_size = ct->size;
	Transform3D* candidates = ct->candidates;

	/* Freeing just the pointer, data are safe in candidates variable */
	free(ct);

	return candidates;
}
