
#include "distance_tree.h"


DistanceTreeNode* distance_tree_node_create() {
	DistanceTreeNode* node = malloc(sizeof(DistanceTreeNode));

	node->type = NON_LEAF;
	node->candidate_index = 0;
	node->close = NULL;
	node->far = NULL;
	node->children = NULL;
	node->children_size = 0;
	node->threshold = 0;

	return node;
}

int distance_tree_compare_float(const void* a, const void* b) {
	float pra = *((float*) a);
	float prb = *((float*) b);

	if(pra == prb) {
		return 0;
	}

	return pra > prb ? 1 : -1;
}

float distance_tree_find_threshold(DistanceTree* tree, int candidate_index, int* children, int children_size) {
	float* distances = malloc(sizeof(float) * children_size);
	for(int i = 0; i < children_size; ++i) {
		distances[i] = metric_compute(tree->candidates[candidate_index], tree->candidates[children[i]]);
	}

	qsort(distances, children_size, sizeof(float), distance_tree_compare_float);
	float median = distances[children_size / 2];

	free(distances);

	return median;
}


DistanceTreeNode* distance_tree_build(DistanceTree* tree, int candidate_index, int* children, int children_size) {
	DistanceTreeNode* node = distance_tree_node_create();

	node->candidate_index = candidate_index;

	/* In case there are less than 5 children. Mark node as leaf, add children and return */
	if(children_size < DISTANCE_TREE_LEAF_CHILDREN_COUNT) {
		node->type = LEAF;
		node->children = malloc(sizeof(int) * children_size);
		memcpy(node->children, children, sizeof(int) * children_size);
		node->children_size = children_size;

		return node;
	}
	
	float threshold = distance_tree_find_threshold(tree, candidate_index, children, children_size);
	node->threshold = threshold;

	VECTOR_CREATE(int) close;
	VECTOR_CREATE(int) far;
	VECTOR_INIT(close, int);
	VECTOR_INIT(far, int);

	/* TODO: Potentional optimalization - those values are already computed in distance_tree_find_threshold. */
	for(int i = 0; i < children_size; ++i) {
		if(metric_compute(tree->candidates[candidate_index], tree->candidates[children[i]]) >= threshold) {
			VECTOR_ADD(far, children[i]);
		} else {
			VECTOR_ADD(close, children[i]);
		}
	}

	if(close.size > 0) {
		/* Remove last candidate from array, and make node out of it */
		int close_index = close.data[close.size - 1];
		node->close = distance_tree_build(tree, close_index, close.data, close.size - 1);
	}
	int far_index = far.data[far.size - 1];
	node->far = distance_tree_build(tree, far_index, far.data, far.size - 1);

	VECTOR_FREE(close);
	VECTOR_FREE(far);

	return node;
}

void distance_tree_find_close_rec(DistanceTreeNode* node, DistanceTree* tree, Transform3D candidate, float threshold, VectorInt* vector) {
	/* If node is a leaf add all children satisfying the condition of being max 'threshold' distance away and return */
	float distance = metric_compute(candidate, tree->candidates[node->candidate_index]);
	if(distance <= threshold) {
		vector_int_add(vector, node->candidate_index);
	}

	if(node->type == LEAF) {
		for(int i = 0; i < node->children_size; ++i) {
			if(metric_compute(candidate, tree->candidates[node->children[i]]) <= threshold) {
				vector_int_add(vector, node->children[i]);
			}
		}
		return;
	}

	if((distance <= (node->threshold + threshold)) && node->close != NULL) {
		distance_tree_find_close_rec(node->close, tree, candidate, threshold, vector);
	}
	if((distance >= (node->threshold - threshold)) && node->far != NULL) {
		distance_tree_find_close_rec(node->far, tree, candidate, threshold, vector);
	}
}

int* distance_tree_find_close(DistanceTree* tree, Transform3D candidate, float threshold, int* result_size) {
	VectorInt* vector = vector_int_create();
	distance_tree_find_close_rec(tree->root, tree, candidate, threshold, vector);

	*result_size = vector->size;
	int* result = vector->data;

	/* Freeing just the variable, keeping the array as it's being returned */
	free(vector);

	return result;
}

DistanceTree* distance_tree_create(Transform3D* candidates, int candidates_size) {
	DistanceTree* tree = malloc(sizeof(DistanceTree));
	tree->candidates = candidates;
	tree->candidates_size = candidates_size;

	int* candidates_indexes = malloc(sizeof(int) * (candidates_size - 1));
	/* Iterating just to "size - 1" because 0 is reserved for root */
	for(int i = 0; i < candidates_size - 1; ++i) {
		candidates_indexes[i] = i + 1;
	}

	tree->root = distance_tree_build(tree, 0, candidates_indexes, candidates_size - 1);

	free(candidates_indexes);

	return tree;
}

void distance_tree_free_rec(DistanceTreeNode* node) {
	if(node->close != NULL) {
		distance_tree_free_rec(node->close);
	}

	if(node->far != NULL) {
		distance_tree_free_rec(node->far);
	}

	if(node->type == LEAF) {
		free(node->children);
	}

	free(node);
}

void distance_tree_free(DistanceTree* tree) {
	distance_tree_free_rec(tree->root);
	free(tree);
}
