
#include "metric.h"

/* Computing the distance of two transformations also involves step
 * which is the same for every pair. Therefore precompute this only once and then
 * just refer to its value, which is, after metric_data_init is called, stored 
 * in this array */
static float data[10];

static float triangle_area_sum;
static float weighted_object_size;
static float object_size;
static float rotation_coeff;

void metric_data_init(TriangleMesh* mesh, float _triangle_area_sum, float _weighted_object_size, float _object_size, float _rotation_coeff) {

	triangle_area_sum = _triangle_area_sum;
	object_size = _object_size;
	rotation_coeff = _rotation_coeff;
	weighted_object_size = _weighted_object_size;

	float L_data[] = {
		1 / sqrt(2), sqrt(2) / 6, sqrt(2) / 6,
		          0,   1.0 / 6.0, -1.0 / 12.0,
				  0,           0, sqrt(3) / 12 
	};

	for(int i = 0; i < mesh->triangles_size; ++i) {
		Triangle t = mesh->triangles[i];

		Point3D x0 = mesh->points[t.v1];
		Point3D b1 = POINTS_SUB(mesh->points[t.v2], x0);
		Point3D b2 = POINTS_SUB(mesh->points[t.v3], x0);

		Point3D qx = create_point3d(x0.x, b1.x, b2.x);
		Point3D qy = create_point3d(x0.y, b1.y, b2.y);
		Point3D qz = create_point3d(x0.z, b1.z, b2.z);

		Point3D Lqx = matrix3x3_vector_multiply_data(L_data, qx);
		Point3D Lqy = matrix3x3_vector_multiply_data(L_data, qy);
		Point3D Lqz = matrix3x3_vector_multiply_data(L_data, qz);
		Point3D m = create_point3d(sqrt(2) / 2, 0, 0);

		Matrix4x4 Q = matrix4x4_create();
		matrix4x4_put_to_row(&Q, Lqx, 0);
		matrix4x4_put_to_row(&Q, Lqy, 1);
		matrix4x4_put_to_row(&Q, Lqz, 2);
		matrix4x4_put_to_row(&Q, m, 3);

		Matrix4x4 QQT = matrix4x4_matrices_multiplication(Q, matrix4x4_transpose(Q));

		float jacobian_det = POINT_ABS(POINTS_CROSSPRODUCT(b1, b2));

		data[0] += jacobian_det * QQT.data[0];
		data[1] += jacobian_det * QQT.data[1];
		data[2] += jacobian_det * QQT.data[2];
		data[3] += jacobian_det * QQT.data[3];
		data[4] += jacobian_det * QQT.data[5];
		data[5] += jacobian_det * QQT.data[6];
		data[6] += jacobian_det * QQT.data[7];
		data[7] += jacobian_det * QQT.data[10];
		data[8] += jacobian_det * QQT.data[11];
		data[9] += jacobian_det * QQT.data[15];
	}
}

float metric_compute_squared(Transform3D c1, Transform3D c2) {

	float a00 = c1.rotation[0] - c2.rotation[0];
	float a01 = c1.rotation[1] - c2.rotation[1];
	float a02 = c1.rotation[2] - c2.rotation[2];
	float a10 = c1.rotation[3] - c2.rotation[3];
	float a11 = c1.rotation[4] - c2.rotation[4];
	float a12 = c1.rotation[5] - c2.rotation[5];
	float a20 = c1.rotation[6] - c2.rotation[6];
	float a21 = c1.rotation[7] - c2.rotation[7];
	float a22 = c1.rotation[8] - c2.rotation[8];
	float a03 = c1.translation[0] - c2.translation[0];
	float a13 = c1.translation[1] - c2.translation[1];
	float a23 = c1.translation[2] - c2.translation[2];

	float result = 0;

	result += a00 * a00 * data[0] + a01 * a01 * data[4] + a02 * a02 * data[7] + a03 * a03 * data[9]
		+ 2 * a00 * a01 * data[1] + 2 * a00 * a02 * data[2] + 2 * a00 * a03 * data[3]
		+ 2 * a01 * a02 * data[5] + 2 * a01 * a03 * data[6]
		+ 2 * a02 * a03 * data[8];
	result += a10 * a10 * data[0] + a11 * a11 * data[4] + a12 * a12 * data[7] + a13 * a13 * data[9]
		+ 2 * a10 * a11 * data[1] + 2 * a10 * a12 * data[2] + 2 * a10 * a13 * data[3]
		+ 2 * a11 * a12 * data[5] + 2 * a11 * a13 * data[6]
		+ 2 * a12 * a13 * data[8];
	result += a20 * a20 * data[0] + a21 * a21 * data[4] + a22 * a22 * data[7] + a23 * a23 * data[9]
		+ 2 * a20 * a21 * data[1] + 2 * a20 * a22 * data[2] + 2 * a20 * a23 * data[3]
		+ 2 * a21 * a22 * data[5] + 2 * a21 * a23 * data[6]
		+ 2 * a22 * a23 * data[8];


	result /= triangle_area_sum;
	result /= weighted_object_size * weighted_object_size;
	result *= object_size * object_size;
	result *= rotation_coeff * rotation_coeff;

	return result;
}

float metric_compute(Transform3D c1, Transform3D c2) {
	return sqrt(metric_compute_squared(c1, c2));
}
