#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "trianglemesh.h"
#include "point3d.h"
#include "triangle.h"
#include "vector.h"

TriangleMesh* get_trianglemesh_from_file(char* filename) {
	FILE* file = fopen(filename, "r");

	if(file == NULL) {
		fprintf(stderr, "%s - Invalid file\n", filename);
		return NULL;
	}

	VECTOR_CREATE(Point3D) points;
	VECTOR_CREATE(Triangle) triangles;
	VECTOR_INIT(points, Point3D);
	VECTOR_INIT(triangles, Triangle);

#define BUFFER_LEN 1024
	char buffer[BUFFER_LEN];
	while(fgets(buffer, BUFFER_LEN, file)) {

		/* Vertex */
		if(buffer[0] == 'v') {
			/* Normals, textures or any other vertex parameters
			 * aren't supported nor needed. Therefore any line with prefix 'vX' 
			 * will get skipped. */
			if(buffer[1] != ' ') {
				continue;
			}

			float x, y, z;
			if(sscanf(buffer, "v %f %f %f", &x, &y, &z) != 3) {
				/* Vertex couldn't be read - error. */
				VECTOR_FREE(points);
				VECTOR_FREE(triangles);
				return NULL;
			}

			Point3D p = create_point3d(x, y, z);
			VECTOR_ADD(points, p);
		}

		/* Triangle */
		if(buffer[0] == 'f') {

			int v1, v2, v3;

			/* Possible face formats from wiki:
			 *   # Polygonal face element (see below)
			 *   f 1 2 3
			 *   f 3/1 4/2 5/3
			 *   f 6/4/1 3/5/3 7/6/5
			 *   f 7//1 8//2 9//3 
			 */
			static const char* possible_formats[] = {"f %d %d %d", "f %d/%*d %d/%*d %d/%*d", "f %d/%*d/%*d %d/%*d/%*d %d/%*d/%*d", "f %d//%*d %d//%*d %d//%*d"};
			char success = 0;

			/* Loops through all possible formats */
			for(int i = 0; i < sizeof(possible_formats) / sizeof(char*); ++i) {
				/* All 3 variables needs to be filled in order to successfully parse
				 * face */
				if(sscanf(buffer, possible_formats[i], &v1, &v2, &v3) == 3) {
					/* .obj format indexes vertexes from 1, therefore for each vertex
					 * subtract 1 to force indexing from 0 */
					v1 -= 1;
					v2 -= 1;
					v3 -= 1;

					success = 1;

					Triangle t = create_triangle(v1, v2, v3);
					VECTOR_ADD(triangles, t);

					/* Triangle was parsed - no need to test for other formats */
					break;
				}
			}
			if(!success) {
				/* Even though line started with 'f' any of the acceptable formats
				 * wasn't recognized - error.*/
				VECTOR_FREE(triangles);
				VECTOR_FREE(points);
				return NULL;
			}
		}
	}

	fclose(file);
	/* Create trianglemesh from obtained data and return it */
	return create_trianglemesh(points.data, points.size, triangles.data, triangles.size);
}
