#include <math.h>
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include "proximity_verifier.h"
#include "point3d.h"
#include "math_utils.h"
#include "config.h"
#include "sampler.h"

void* print_kernel(void* arguments) {
	T_Arguments* ta =  (T_Arguments*) arguments;
	T_Arguments_Static args = *(ta->args);
	
	int start = ta->start;
	int n = ta->to_process;
	char* kernel = args.kernel;
	char* grid = args.grid;
	int kernel_span = args.kernel_span;

	if(start + n >= args.points_size) {
		n -= (start + n) - args.points_size;
	}

	for(int i = start; i < start + n; ++i) {
		Point3D p = (args.points)[i];
		int x = (int) ((p.x - args.min.x) / args.step);
		int y = (int) ((p.y - args.min.y) / args.step);
		int z = (int) ((p.z - args.min.z) / args.step);

		//printf("Point %d corresponds to indexes [%d; %d; %d]\n", i, x, y, z);

		int d_span = args.d_span;
		int dspan = args.d_span;

		int xpd = -x + dspan;
		int ypd = -y + dspan;
		int zpd = -z + dspan;

		for (int lx = x - d_span; lx < x + d_span; ++lx) {
			if(lx < 0 || lx >= args.grid_x) {
				continue;
			}
			for (int ly = y - d_span; ly < y + d_span; ++ly) {
				if(ly < 0 || ly >= args.grid_y) {
					continue;
				}

				for (int lz = z - d_span; lz < z + d_span; ++lz) {
					if(lz < 0 || lz >= args.grid_z) {
						continue;
					}

					/* TODO: DOes this use cache well? Either lz should be the last term
					 * or the very inner iterating (over lz now) should be lx instead. */
					int index = args.grid_x * args.grid_y * lz + args.grid_x * ly + lx;

					/* Maps lx ly lz to kernel space, that's <0; d_span*2 + 1> */
					if(kernel[(lx - x + d_span) * (kernel_span * kernel_span) + (ly - y + d_span) * kernel_span + (lz - z + d_span)] == 1) {
						/* TODO: Move calculating index here. Maybe next step whould be
						 * checking if grid[index] is 1 first and then check for kernel. */
						grid[index] = 1;
					}
				}
			}
		}
	}
}

/* Creates 3D 'bitmap' sphere with side being kernel_span long */
char* build_kernel(int d_span, float threshold, int kernel_span) {
	char* kernel = calloc(kernel_span * kernel_span * kernel_span, sizeof(char));

	for (int lx = 0; lx < kernel_span; lx++) {
		for (int ly = 0; ly < kernel_span; ly++) {
			for (int lz = 0; lz < kernel_span; lz++) {

				float ds = (lx - d_span) * (lx - d_span) + (ly - d_span) * (ly - d_span) + (lz - d_span) * (lz - d_span);

				if (ds < threshold) {
					kernel[lx * kernel_span * kernel_span + ly * kernel_span + lz] = 1;
				}
			}
		}
	}
	return kernel;
}

void start_workers(T_Arguments_Static tas, int num, int to_process) {

	pthread_t tid;
	pthread_t* tids = malloc(sizeof(pthread_t) * num);
	T_Arguments** t_arg_pointers = (T_Arguments**) malloc(sizeof(T_Arguments*) * num);

	for(int i = 0; i < num; ++i) {
		T_Arguments* ta = malloc(sizeof(T_Arguments));
		ta->start = i * to_process;
		ta->args = &tas;
		ta->to_process = to_process;

		pthread_create(&tid, NULL, print_kernel, (void*) ta);

		tids[i] = tid;
		t_arg_pointers[i] = ta;
	}

	for(int i = 0; i < num; ++i) {
		pthread_join(tids[i], NULL);
	}

	for(int i = 0; i < num; ++i) {
		free(t_arg_pointers[i]);
	}

	free(tids);
	free(t_arg_pointers);
}

int proximity_verifier_verify(const ProximityVerifier* verifier, Transform3D transform) {

	/* Count of points, for which there are points near after transformation. */
	int good_points = 0;

	/* Transfrom each point from mesh2 and check. If there's point from mesh1 in proximity.
	 * (There's '1' present in corresponding grid cell) */
	for(int i = 0; i < verifier->mesh2_size; ++i) {
		Point3D p = verifier->mesh2[i];

#ifdef DEBUG
		printf("BEFORE: ");
		print_point3d(p);
#endif
		/* Transforming the point */
		p = transform3d_transform_point(transform, p);
#ifdef DEBUG
		printf("AFTER: ");
		print_point3d(p);
#endif

		int ix = (int) ((p.x - verifier->min_point.x) * verifier->inverse_step);
		int iy = (int) ((p.y - verifier->min_point.y) * verifier->inverse_step);
		int iz = (int) ((p.z - verifier->min_point.z) * verifier->inverse_step);

#ifdef DEBUG
		printf("Corresponding indexes: [%d; %d; %d]\n", ix, iy, iz);
#endif

		/* Check for boundaries of grid. Transformed point might be outside the grid */
		if(ix < 0 || ix > verifier->grid_x) {
			continue;
		}
		if(iy < 0 || iy > verifier->grid_y) {
			continue;
		}
		if(iz < 0 || iz > verifier->grid_z) {
			continue;
		}

		/* If the point lies within the grid, check for grid value. */
		if(verifier->grid[verifier->grid_x * verifier->grid_y * iz + verifier->grid_x * iy + ix]) {
			good_points += 1;
		}

		/* If sufficient number of transformed points lie in proximity of mesh1 return 1*/
		if(good_points >= verifier->points_acceptance_threshold) {
			return 1;
		}
	}

	return 0;
}

void proximity_verifier_sample_mashes(TriangleMesh* mesh1, TriangleMesh* mesh2, Point3D** sampled_mesh1, int* sampled_mesh1_size, Point3D** sampled_mesh2, int* sampled_mesh2_size) {
	int* sampled_mesh2_indexes;

	//sampler_uniform_sampling_grid(mesh1, config_sample_size_mesh1, (mesh2->radius * config_delta_ratio) / 3.0, &sampled_mesh1_indexes, sampled_mesh1_size);
	sampler_sample(mesh2, config_sample_size_mesh2, &sampled_mesh2_indexes, sampled_mesh2_size);

	*sampled_mesh1_size = mesh1->points_size;
	*sampled_mesh1 = malloc(sizeof(Point3D) * (mesh1->points_size));
	*sampled_mesh2 = malloc(sizeof(Point3D) * (*sampled_mesh2_size));

	/* Transform indexes to points. */
	//for(int i = 0; i < *sampled_mesh1_size; ++i) {
	//	(*sampled_mesh1)[i] = mesh1->points[sampled_mesh1_indexes[i]];
	//}

	for(int i = 0; i < mesh1->points_size; ++i) {
		(*sampled_mesh1)[i] = mesh1->points[i];
	}
	for(int i = 0; i < *sampled_mesh2_size; ++i) {
		(*sampled_mesh2)[i] = mesh2->points[sampled_mesh2_indexes[i]];
	}

	free(sampled_mesh2_indexes);
}

ProximityVerifier* create_proximity_verifier(TriangleMesh* mesh1, TriangleMesh* mesh2, float delta) {
	/* Again, for the sake of performance sample meshes */
	Point3D* sampled_mesh1;
	int sampled_mesh1_size;
	Point3D* sampled_mesh2;
	int sampled_mesh2_size;

	proximity_verifier_sample_mashes(mesh1, mesh2, &sampled_mesh1, &sampled_mesh1_size, &sampled_mesh2, &sampled_mesh2_size);

	ProximityVerifier* verifier = malloc(sizeof(ProximityVerifier));

	if(verifier == NULL) {
		return NULL;
	}

	verifier->mesh1 = sampled_mesh1;
	verifier->mesh1_size = sampled_mesh1_size;

	verifier->mesh2 = sampled_mesh2;
	verifier->mesh2_size = sampled_mesh2_size;

	verifier->points_acceptance_threshold = ceilf(config_acceptace_threshold * sampled_mesh2_size);

	/* First - finding bounding box */
	Point3D min_p = sampled_mesh1[0];
	Point3D max_p = sampled_mesh1[0];

	for(int i = 0; i < sampled_mesh1_size; ++i) {
		if(sampled_mesh1[i].x < min_p.x) {
			min_p.x = sampled_mesh1[i].x;
		}
		if(sampled_mesh1[i].y < min_p.y) {
			min_p.y = sampled_mesh1[i].y;
		}
		if(sampled_mesh1[i].z < min_p.z) {
			min_p.z = sampled_mesh1[i].z;
		}

		if(sampled_mesh1[i].x > max_p.x) {
			max_p.x = sampled_mesh1[i].x;
		}
		if(sampled_mesh1[i].y > max_p.y) {
			max_p.y = sampled_mesh1[i].y;
		}
		if(sampled_mesh1[i].z > max_p.z) {
			max_p.z = sampled_mesh1[i].z;
		}
	}

	verifier->min_point = min_p;

	/* max_p now is difference between pervious max_p and min_p */
	POINTS_SUB_REF(&max_p, min_p);

	float max_span = fmax(fmax(max_p.x, max_p.y), max_p.z);
	float step = max_span / NUMBER_OF_CELLS;

	verifier->step = step;
	verifier->inverse_step = 1.0 / step;

	int grid_x = (int) (max_p.x / step) + 1;
	int grid_y = (int) (max_p.y / step) + 1;
	int grid_z = (int) (max_p.z / step) + 1;

	char* grid = calloc(grid_x * grid_y * grid_z, sizeof(char));

	verifier->grid_x = grid_x;
	verifier->grid_y = grid_y;
	verifier->grid_z = grid_z;
	verifier->grid = grid;

	if(verifier->grid == NULL) {
		/* Instead of allocating grid_x * grid_y * grid_z try allocating 
		 * 8 times less space and do bit array emulation, but I don't think
		 * it will ever be needed */
		return NULL;
	}

	/* How many cells around cell with a point should still be considered
	 * a match(Only used for very small steps)  */
	int d_span = (int) ceil(delta / step);

	int kernel_span = 2 * d_span + 1;
	float threshold = (delta / step) * (delta / step);

	char* kernel = build_kernel(d_span, threshold, kernel_span);


	T_Arguments_Static tas = {sampled_mesh1, sampled_mesh1_size, grid, grid_x, grid_y, grid_z, min_p, kernel, d_span, kernel_span, step};
	int number_of_processors = config_number_of_processors;
	int to_process = ceil((float) sampled_mesh1_size / number_of_processors);

	start_workers(tas, number_of_processors, to_process);

	free(kernel);

	return verifier;
}

void proximity_verifier_free(ProximityVerifier* verifier) {
	free(verifier->mesh1);
	free(verifier->mesh2);
	free(verifier->grid);
	free(verifier);
}
